@define-color documents_emblem_bg #3465a4;
@define-color documents_collection_bg #d3d7cf;
@define-color theme_base_color #ffffff;
@define-color theme_bg_color #ededed;
@define-color theme_selected_bg_color #4a90d9;
@define-color borders #a7aba7;

GdMainIconView.content-view {
    -GdMainIconView-icon-size: 40;
}

/* Make spinner visible on both dark and bright backgrounds w/o making
 * it look ugly/weird.
 */
GdMainIconView.content-view.cell:active {
    color: gray;
}

.documents-load-more.button {
    border-image: none;
    border-color: @borders;
    border-width: 1px 0 0;
    border-radius: 0;
}

.documents-scrolledwin.frame {
    border-width: 1px 0 0;
    border-radius: 0;
}

.documents-icon-bg {
    background-color: @documents_emblem_bg;
    border-radius: 4px;

    color: @theme_base_color;
}

.documents-collection-icon {
    background-color: @documents_collection_bg;
    border-radius: 8px;
}

.documents-counter {
    background-image: url("ui/dnd-counter.svg");
    background-size: contain;
    background-color: transparent;
    color: @theme_base_color;
    font: bold;
}

.documents-favorite.button:active,
.documents-favorite.button:active:hover {
    color: shade(@theme_selected_bg_color, 1.20);
}

.content-view.document-page {
    border-style: solid;
    border-width: 3px 3px 6px 4px;
    border-image: url("ui/thumbnail-frame.png") 3 3 6 4;
}

.osd .page-thumbnail {
    border-style: solid;
    border-width: 1px;
    border-color: @osd_lowlight;

    /* when there's no pixbuf yet */
    background-color: @osd_bg;
}

.collection-dialog .frame { border-style: solid none none none; }

.collection-dialog .delete-row { background-color: @theme_bg_color }

.collection-dialog .frame GtkEntry {
  padding: 12px 8px 13px;
  border-left: none;
  border-right: none;
  border-top: none;
  border-bottom: none;
  border-radius: 0px;
}
